/**
 * Copyright 2014 Daum Kakao Corp.
 *
 * Redistribution and modification in source or binary forms are not permitted without specific prior written permission. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.kakao.auth.api;

import android.content.Context;
import android.net.Uri;
import android.os.Bundle;
import android.os.Handler;
import android.os.Looper;
import android.webkit.CookieSyncManager;

import com.kakao.auth.AuthService.AgeAuthLevel;
import com.kakao.auth.AuthService.AgeAuthStatus;
import com.kakao.auth.AuthService.AgeLimit;
import com.kakao.auth.KakaoSDK;
import com.kakao.auth.Session;
import com.kakao.auth.StringSet;
import com.kakao.auth.authorization.AuthorizationResult;
import com.kakao.auth.authorization.accesstoken.AccessToken;
import com.kakao.auth.authorization.accesstoken.AccessTokenRequest;
import com.kakao.auth.authorization.authcode.KakaoWebViewDialog;
import com.kakao.auth.authorization.authcode.KakaoWebViewDialog.OnWebViewCompleteListener;
import com.kakao.network.ServerProtocol;
import com.kakao.auth.SingleNetworkTask;
import com.kakao.auth.network.request.AccessTokenInfoRequest;
import com.kakao.auth.network.response.AccessTokenInfoResponse;
import com.kakao.network.response.ResponseBody;
import com.kakao.network.response.ResponseData;
import com.kakao.util.exception.KakaoException;
import com.kakao.util.helper.Utility;
import com.kakao.util.helper.log.Logger;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Bloking으로 동작하며, 인증관련 내부 API콜을 한다.
 * @author leoshin
 */
public class AuthApi {
    public static void synchronizeCookies(Context context) {
        CookieSyncManager syncManager = CookieSyncManager.createInstance(context);
        syncManager.sync();
    }

    public static AuthorizationResult requestAccessToken(Context context, String appKey, String redirectUri, String authCode, String refreshToken, String approvalType) throws Exception {
        SingleNetworkTask networkTask = new SingleNetworkTask();
        ResponseBody result = networkTask.requestAuth(new AccessTokenRequest(context, appKey, redirectUri, authCode, refreshToken, approvalType));
        final AccessToken accessToken = new AccessToken(result);
        return AuthorizationResult.createSuccessAccessTokenResult(accessToken);
    }

    private static boolean requestWebviewAuth(Context context, String accessToken, String redirectUri, AgeAuthLevel authLevel, AgeLimit ageLimit, String ageauthMode, OnWebViewCompleteListener listener) {
        try {
            final Bundle parameters = new Bundle();
            parameters.putString(StringSet.token_type, "api");
            parameters.putString(StringSet.access_token, accessToken);
            parameters.putString(StringSet.return_url, redirectUri);

            if (authLevel != null) {
                parameters.putString(StringSet.ageauth_level, authLevel.getValue());
            }
            if (ageLimit != null) {
                parameters.putString(StringSet.age_limit, ageLimit.getValue());
            }
            if (ageauthMode != null && ageauthMode.length() > 0) {
                parameters.putString(StringSet.ageauth_mode, ageauthMode);
            }

            synchronizeCookies(context);

            Uri uri = Utility.buildUri(ServerProtocol.AGE_AUTH_AUTHORITY, ServerProtocol.ACCESS_AGE_AUTH_PATH, parameters);
            KakaoWebViewDialog loginDialog = new KakaoWebViewDialog(context, uri.toString(), KakaoSDK.getAdapter().getSessionConfig().isUsingWebviewTimer(), listener);
            loginDialog.show();
        } catch (Throwable t) {
            Logger.e("WebViewAuthHandler is failed", t);
            return false;
        }
        return true;
    }

    public static int requestShowAgeAuthDialog(final Context context, final AgeAuthLevel authLevel, final AgeLimit ageLimit, final String ageauthMode) {
        final AtomicInteger result = new AtomicInteger();
        final CountDownLatch lock = new CountDownLatch(1);
        new Handler(Looper.getMainLooper()).post(new Runnable() {
            @Override
            public void run() {
                try {
                    OnWebViewCompleteListener listener = new OnWebViewCompleteListener() {
                        @Override
                        public void onComplete(String redirectURL, KakaoException exception) {
                            int status = AgeAuthStatus.CLIENT_ERROR.getValue();
                            if (redirectURL != null) {
                                if (Uri.parse(redirectURL).getQueryParameter(StringSet.status) != null) {
                                    status = Integer.valueOf(Uri.parse(redirectURL).getQueryParameter(StringSet.status));
                                }
                            }

                            result.set(status);
                            lock.countDown();
                        }
                    };

                    String redirectUri = Session.getCurrentSession().generateRedirectUrl();
                    String accessToken = Session.getCurrentSession().getAccessToken();
                    requestWebviewAuth(context, accessToken, redirectUri, authLevel, ageLimit, ageauthMode, listener);
                } catch (Exception e) {
                    result.set(AgeAuthStatus.CLIENT_ERROR.getValue());
                    lock.countDown();
                }
            }
        });

        // 사용자가 취소를 하여도 종료.
        try {
            lock.await();
        } catch (InterruptedException ignor) {
        }

        return result.get();
    }

    public static int requestAccessTokenInfo() throws Exception {
        SingleNetworkTask networkTask = new SingleNetworkTask();
        ResponseData result = networkTask.requestApi(new AccessTokenInfoRequest());
        return new AccessTokenInfoResponse(result).getExpiresInMillis();
    }
}
